
mutable struct ForestIterator
    # how many leaves do the left subtrees have
    left_leaves::Int
    left_subiterator::Union{ForestIterator,Missing}

    # how many leaves do the right subtrees have
    right_leaves::Int
    right_subiterator::Union{ForestIterator,Missing}

    # counter is like an id and incremented each time _iterate is called
    # important for symmetry breaking for iterators where n is even and the subtrees can have equal leave numbers
    counter::Int
end

function ForestIterator(n::Int)
    @assert n >= 1 "Cannot create a ForestIterator with fewer than 1 leaves"
    if n == 1
        return missing
    end
    return ForestIterator(1, ForestIterator(1), n - 1, ForestIterator(n - 1), 1)
end

ForestIterator(forest::Forest) = ForestIterator(forest.leaves)

"""
    IteratorSize(::Forest)

Would be OEIS sequence A001190, Wedderburn-Etherington numbers: Number of unlabeled binary rooted trees with n leaves.
"""
Base.IteratorSize(::Forest) = Base.SizeUnknown()
Base.IteratorSize(::ForestIterator) = Base.SizeUnknown()

"""
    is_end(::ForestIterator)
    
Return whether the given [`ForestIterator`](@ref) is exhausted.
"""
is_end(::Missing) = true
is_end(iterator::ForestIterator) = iterator.left_leaves + 1 >= iterator.right_leaves && is_end(iterator.left_subiterator) && is_end(iterator.right_subiterator)

"""
    tree(iterator::ForestIterator)

Return an [`AbstractTree`](@ref) made from the current state of the iterator.
"""
tree(::Missing) = EmptyTree()
function tree(iterator::ForestIterator)
    return Subtree(tree(iterator.left_subiterator), tree(iterator.right_subiterator))
end

"""
    _iterate(iterator::ForestIterator)

Internal function for iterating the [`ForestIterator`](@ref) once.
"""
function _iterate(iterator::ForestIterator)
    @assert !is_end(iterator) "Trying to iterate a forest iterator that is already at its end"
    iterator.counter += 1
    # reverse direction, iterate right first, then left (reset right), then try new balance
    if !is_end(iterator.right_subiterator)
        _iterate(iterator.right_subiterator)
        return nothing
    elseif !is_end(iterator.left_subiterator)
        iterator.right_subiterator = ForestIterator(iterator.right_leaves)

        if (iterator.left_leaves == iterator.right_leaves)
            # if the subiterators have equal lengths, take care to start the one we reset where the other one currently is (to not get duplicates)
            # this actually only changes things at forest(8) and higher
            while (iterator.right_subiterator.counter <= iterator.left_subiterator.counter)
                _iterate(iterator.right_subiterator)
            end
        end
        _iterate(iterator.left_subiterator)

        return nothing
    else
        iterator.right_leaves -= 1
        iterator.left_leaves += 1
        iterator.right_subiterator = ForestIterator(iterator.right_leaves)
        iterator.left_subiterator = ForestIterator(iterator.left_leaves)
        return nothing
    end
end

function Base.iterate(forest::Forest)
    state = ForestIterator(forest.leaves)
    return (tree(state), state)
end

function Base.iterate(forest::Forest, state::Missing)
    return nothing
end

function Base.iterate(forest::Forest, state::ForestIterator)
    if (is_end(state))
        return nothing
    end

    _iterate(state)
    return (tree(state), state)
end
